#include <stdio.h>
#include <stdbool.h>
#include "app_error.h"
#include "app_uart.h"
#include "nrf_delay.h"

#include "nrf_drv_saadc.h"
#include "nrf_drv_ppi.h"
#include "nrf.h"
#include "nrf_drv_timer.h"

#include "app_util_platform.h"
#include "nrf_delay.h"
#include "bsp.h"
#include "app_timer.h" 
#include "nrf_drv_spi.h"
#include "nordic_common.h"
#include "SEGGER_RTT.h"
#include "nrf_soc.h"

#include "config.h"
#include "nrf_drv_gpiote.h"

#include "softdevice_handler.h"
#include "myheader.h"

nrf_ppi_channel_t ppi_channel_workaround;

//setting up the SPI based on the spi_master_with_spi_slave example
//uses DMA
#if (SPI0_ENABLED == 1)
    static const nrf_drv_spi_t m_spi_master = NRF_DRV_SPI_INSTANCE(0);

#else
    #error "No SPI enabled."
#endif

static volatile bool m_transfer_completed = true; /**< A flag to inform about completed transfer. */

/**@brief Function for SPI master event callback.
 *
 * Upon receiving an SPI transaction complete event, checks if received data are valid.
 *
 * @param[in] spi_master_evt    SPI master driver event. 
 */
static void spi_master_event_handler(nrf_drv_spi_event_t event)
{
    uint32_t err_code = NRF_SUCCESS;
    switch (event)
    {
        
			  case NRF_DRV_SPI_EVENT_DONE:   
						
						//disable the ppi channel
            err_code = nrf_drv_ppi_channel_disable(ppi_channel_workaround);
						
						//disable a GPIOTE output pin task
						nrf_drv_gpiote_out_task_disable(SPIM0_SCK_PIN);
				
						// Check if data are valid.
            err_code = bsp_indication_set(BSP_INDICATE_RCV_OK);
            APP_ERROR_CHECK(err_code);

            // Inform application that transfer is completed.
            m_transfer_completed = true;
            break;
				
				//if event is not equal to NRF_DRV_SPI_EVENT_DONE, then nothing happens
        default:
            // No implementation needed.
            break;
    }
}


/**@brief Function for initializing bsp module.
 */
void bsp_configuration()
{
		SEGGER_RTT_WriteString(0, "BSP Started\n");
    uint32_t err_code = NRF_SUCCESS;

    NRF_CLOCK->LFCLKSRC            = (CLOCK_LFCLKSRC_SRC_Xtal << CLOCK_LFCLKSRC_SRC_Pos);
    NRF_CLOCK->EVENTS_LFCLKSTARTED = 0;
    NRF_CLOCK->TASKS_LFCLKSTART    = 1;

    while (NRF_CLOCK->EVENTS_LFCLKSTARTED == 0)
    {
        // Do nothing.
    }
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_MAX_TIMERS, APP_TIMER_OP_QUEUE_SIZE, NULL);
    err_code = bsp_init(BSP_INIT_LED, APP_TIMER_TICKS(100, APP_TIMER_PRESCALER), NULL);
    APP_ERROR_CHECK(err_code);
}

void uart_error_handle(app_uart_evt_t * p_event)
{
    if (p_event->evt_type == APP_UART_COMMUNICATION_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_communication);
    }
    else if (p_event->evt_type == APP_UART_FIFO_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_code);
    }
}

void uart_config(void){
		uint32_t err_code;
		const app_uart_comm_params_t comm_params =
      {
          RX_PIN_NUMBER,
          TX_PIN_NUMBER,
          RTS_PIN_NUMBER,
          CTS_PIN_NUMBER,
          APP_UART_FLOW_CONTROL_ENABLED,
          false,
          UART_BAUDRATE_BAUDRATE_Baud38400
      };

    APP_UART_FIFO_INIT(&comm_params,
                         UART_RX_BUF_SIZE,
                         UART_TX_BUF_SIZE,
                         uart_error_handle,
                         APP_IRQ_PRIORITY_LOW,
                         err_code);
	
		APP_ERROR_CHECK(err_code);
}

/**
 * @brief Work-around for transmitting 1 byte with SPIM.
 *
 * @param spim: The SPIM instance that is in use.
 * @param ppi_channel: An unused PPI channel that will be used by the
		workaround.
 * @param gpiote_channel: An unused GPIOTE channel that will be used by
		the workaround.
 *
 * @warning Must not be used when transmitting multiple bytes.
 * @warning After this workaround is used, the user must reset the PPI
		channel and the GPIOTE channel before attempting to transmit multiple
		bytes.
 */
void setup_workaround_for_ftpan_58(NRF_SPIM_Type *spim, 
																	uint32_t ppi_channel, 
																	uint32_t gpiote_channel)
{
	
	// Create an event when SCK toggles.
	NRF_GPIOTE->CONFIG[gpiote_channel] = 
		(GPIOTE_CONFIG_MODE_Event << GPIOTE_CONFIG_MODE_Pos) | 
		(spim->PSEL.SCK << GPIOTE_CONFIG_PSEL_Pos) | 
		(GPIOTE_CONFIG_POLARITY_Toggle <<GPIOTE_CONFIG_POLARITY_Pos);

	// Stop the spim instance when SCK toggles.
	NRF_PPI->CH[ppi_channel].EEP = (uint32_t)&NRF_GPIOTE->EVENTS_IN[gpiote_channel];
	NRF_PPI->CH[ppi_channel].TEP = (uint32_t)&spim->TASKS_STOP;
	NRF_PPI->CHENSET = 1U << ppi_channel;
}

void in_pin_handler(nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action)
{
    nrf_drv_gpiote_out_toggle(SPIM0_SCK_PIN);
}

/**@brief Functions prepares buffers and starts data transfer to AS3911
 *
 * @param[in] p_tx_data     A pointer to a buffer TX.
 * @param[in] p_rx_data     A pointer to a buffer RX.
 * @param[in] len           A length of the data buffers.
 */
static void spi_send_recv(uint8_t * const p_tx_data,
                          uint8_t * const p_rx_data,
                          const uint16_t  len)
{
		uint32_t err_code;
		NRF_SPIM_Type *p_spim = m_spi_master.p_registers;
		SEGGER_RTT_WriteString(0, "into send/recv \n");
	
    if (len == 1){
			nrf_delay_ms(delay);
			SEGGER_RTT_WriteString(0, "Sending One Byte \n");
			ret_code_t err_code;
			
			//initializes the GPIOTE channel so that SCK toggles generates events
			nrf_drv_gpiote_in_config_t config = GPIOTE_CONFIG_IN_SENSE_TOGGLE(true);
			err_code = nrf_drv_gpiote_in_init(SPIM0_SCK_PIN, &config, in_pin_handler);
			APP_ERROR_CHECK(err_code);	
	
			//allocates the first unused PPI Channel
			err_code = nrf_drv_ppi_channel_alloc(&ppi_channel_workaround);
			APP_ERROR_CHECK(err_code);
			nrf_delay_ms(delay);
			
			//there's an error here!!!!!
			nrf_delay_ms(delay);
			SEGGER_RTT_printf(0, "The event address is %p \n",nrf_drv_gpiote_in_event_addr_get(SPIM0_SCK_PIN));
		  nrf_delay_ms(delay);
			SEGGER_RTT_printf(0, "The task address is %p \n",p_spim->TASKS_STOP); //adress is 0
			err_code = nrf_drv_ppi_channel_assign(ppi_channel_workaround, nrf_drv_gpiote_in_event_addr_get(SPIM0_SCK_PIN), p_spim->TASKS_STOP);
			nrf_delay_ms(delay);
			SEGGER_RTT_printf(0, "The error is %x \n",err_code);
			APP_ERROR_CHECK(err_code);
	
			//enable the PPI channel.
			err_code = nrf_drv_ppi_channel_enable(ppi_channel_workaround);
			APP_ERROR_CHECK(err_code);
	
			//enable the GPIOTE input pin event.
			nrf_drv_gpiote_in_event_enable(SPIM0_SCK_PIN, false);

			//Start transfer of data
			err_code = nrf_drv_spi_transfer(&m_spi_master, p_tx_data, len, p_rx_data, len);
			APP_ERROR_CHECK(err_code);
														
		}
		else{
			SEGGER_RTT_WriteString(0, "Send multiple bytes \n");
			// Start transfer.
			err_code = nrf_drv_spi_transfer(&m_spi_master, p_tx_data, len, p_rx_data, len);
			APP_ERROR_CHECK(err_code);
		}
		
    nrf_delay_ms(delay);
}


/**@brief Function for sending two bytes to AS3911
*
* @param[in] reg is the register of the AS3911 that you want to access
* @param[in] val is the data written to the AS3911 through SPI
*/
void as3911WriteRegister(uint8_t reg, uint8_t val){
	uint8_t write_buffer[2];
	
	write_buffer[0] = reg | AS3911_WRITE_MODE;
	write_buffer[1] = val;
	spi_send_recv(write_buffer, NULL, 2);
	nrf_delay_ms(delay);
	SEGGER_RTT_printf(0, "Master sent %x, %x \n", write_buffer[0], write_buffer[1]);
}

/**@brief Function for sending 1 byte to AS3911
*
* @param[in] cmd is the command you want to send to the AS3911 
*/
void as3911ExecuteCommand(uint8_t cmd){
	uint8_t write_buffer[1];
	write_buffer[0] = cmd | AS3911_CMD_MODE; //11XX XXXX
	nrf_delay_ms(delay);
	SEGGER_RTT_WriteString(0, "Executing command\n");
	
	spi_send_recv(write_buffer, NULL, 1);
	nrf_delay_ms(delay);
	SEGGER_RTT_printf(0, "Master sent %x\n", write_buffer[0]);
	
}

/**@brief Function for reading bytes from AS3911
*@param[in] register 
*/

int main(void){
		//0 is the terminal number where the text is output on the J-link RTT
		//These lines are meant for debugging purposes and are not necessary 
		nrf_delay_ms(delay);
		SEGGER_RTT_WriteString(0, "Program Begin\n");			
		
		//INITIALIZING PERIPHERALS
		//Initialize UART, and board support
		uart_config();
		bsp_configuration();
		
		//Configure SPI interface
    nrf_drv_spi_config_t const config =
    {
        #if (SPI0_ENABLED == 1)
            .sck_pin  = SPIM0_SCK_PIN, 
            .mosi_pin = SPIM0_MOSI_PIN,
            .miso_pin = SPIM0_MISO_PIN,
            .ss_pin   = SPIM0_SS_PIN,
        
        #endif
						.irq_priority = APP_IRQ_PRIORITY_LOW, //tells computer what to prioritize over other things
						.orc          = 0x00,
						.frequency    = NRF_DRV_SPI_FREQ_1M,
						.mode         = NRF_DRV_SPI_MODE_1,
						.bit_order    = NRF_DRV_SPI_BIT_ORDER_MSB_FIRST,
    };
		
		//Initialize SPI
		ret_code_t err_code = nrf_drv_spi_init(&m_spi_master, &config, spi_master_event_handler);
    APP_ERROR_CHECK(err_code);
		
		//initialize PPI and GPIOTE
    err_code = nrf_drv_ppi_init();
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_gpiote_init();
    APP_ERROR_CHECK(err_code);
		
		//VARIABLES
		uint8_t numBytesReg = 0x1D;
		delay = 100;
		
		//Infinite loop containing main functionality
		for(;;){
		  nrf_delay_ms(delay);
			SEGGER_RTT_WriteString(0, "Enterred For loop \n");	
			
			as3911WriteRegister(numBytesReg, 0xFF);
			as3911ExecuteCommand(0xC1);
			
			if (m_transfer_completed)
			{
				SEGGER_RTT_WriteString(0, "SPI transfer successful \n");			
				m_transfer_completed = false;
			}
			nrf_delay_ms(delay);
		}
}		
